""" Non-Deterministic Buchi Automaton (NBA)

A NFA is the tuple: <states, alphabet, tr, init, final>

WARNING: here we represent the alphabet indirectly with the 
set of atomic propositions AP, so alphabet = 2^AP


"""
from __future__ import annotations

import sys
from collections.abc import MutableMapping, Iterator, Iterable, Generator
from typing import Tuple, Optional, Collection, Set
from functools import reduce, partial

from mc308.logics.bool import BoolFormula
from mc308.auto.ts import TS, State
from mc308.auto.nfa import AutTrans


class NBA():
    def __init__(self : NBA,
                 states : Collection[State],
                 tr : Collection[AutTrans],
                 init : Collection[State],
                 final : Collection[State],
                 ap : Collection[BoolFormula]):
        """
        Creates a finite state automaton
        """

        self.states = set(states)
        self.init = set(init)
        assert(self.init <= self.states)
        self.final = set(final)
        assert(self.final <= self.states)

        self.ap = set(ap)
        assert reduce(lambda x,y : x and y.is_symbol(), self.ap, True)

        self.tr : MutableMapping[State, set[BoolFormula]]= {
            q : set() for q in self.states
        }
        for (s,a,p) in tr:
            assert s in self.states and p in self.states
            aset = frozenset(a)
            assert(aset <= self.ap)

            self.tr[s].add((aset,p))

    def __repr__(self : NBA) :
        pl = lambda l : ",".join(l)

        tr_list = set()
        for s in self.states:
            for (a,b) in self.tr.get(s, []):
                tr_list.add((s,a,b))

        return """NBA : {
  states : %s;
  init : %s;
  final : %s;
  tr : %s;
  ap : %s;
}""" % ( pl([str(l) for l in self.states]),
         pl([str(l) for l in self.init]),
         pl([str(l) for l in self.final]),
         pl(["(%s, {%s}, %s)" % (str(a[0]), pl([str(b) for b in a[1]]), str(a[2])) for a in tr_list]),
         pl([str(a) for a in self.ap]),
        )


