"""
Data structure of a symbolic transition system.

The data structure reprsent a transition system using Propositional logic formulas.

"""

from mc308.logics.bool import BoolFormula

from pysmt.shortcuts import (
    get_env, Symbol, And
)

from pysmt.smtlib.parser import SmtLibParser


from typing import Type, Tuple, Optional, NewType, Collection, Union, Sequence, List, Any


NEXT_SUFFIX = "_next"

def get_next(v):
    """
    Given v, returns v'
    """
    return Symbol("%s%s" % (v.symbol_name(), NEXT_SUFFIX), v.symbol_type())


def get_nexts(V):
    """
    Given a list of variables [v1,v2, ...], returns [v1', v2', ...]
    """
    return [Symbol("%s%s" % (v.symbol_name(),NEXT_SUFFIX), v.symbol_type()) for v in V]


class SymbolicTS():
    def __init__(self, V : Collection[BoolFormula], init : BoolFormula, trans : BoolFormula) -> None:
        """
        Symbolic transition system:
        V: set of Boolean variables (Represented as a Symbol from PySMT)
        I: initial states, formula only containing variable from V (PySMT formula)
        T: transition relation, formula only containing variable from V and V', here called V_next (PySMT formula)

        About V_next:
        We assume we can map a variables from V to V_next just by modifying the name of v.
        Given a variable "a" you get the primed variable named a_next.

        This means that your initial set of variables should not already terminate
        with the suffix _next (otherwise, there could be some variable capture...)

        Given a variable, you can get the corresponding next variable with the function
        get_next.

        """
        self.V = set([v for v in V])
        self.V_next = set([get_next(v) for v in V])
        self.init = init
        self.trans = trans


    def retime(self, formula : BoolFormula,
               time : int,
               time_next : Optional[int] = None) -> BoolFormula :
        """
        Given a formula containing the variables V and V_next and an integer >= 0,
        returns a formula where:
        - all the V-s variables have been replaced with v_at_time variables, and
        - all the V_next-s variables v_next have been replaced with v_at_<time+1>.

        Example
        Assume V = {a,b,c}, V_next = {a_next, b_next, c_next}, and formula is
        the formula "(a & b & c) | !c')".

        self.retime(formula, 2) returns the formula:
        "(a_at_2 & b_at_2 & c_at_2) | !c_at_3)".
        """

        def at_time(v, time, is_next = False):
            assert time >= 0

            if not is_next:
                assert not v.symbol_name().endswith("_next")
                symb_name = "%s_at_%d" % (v.symbol_name(),time) 
            else:
                assert v.symbol_name().endswith("_next")
                name = v.symbol_name()
                name = name[:-len("_next")]
                symb_name = "%s_at_%d" % (name,time)
            return Symbol(symb_name, v.symbol_type())

        time_next = time + 1 if time_next is None else time_next

        f1 = formula.substitute({ a : at_time(a,time) for a in self.V} )
        f2 = f1.substitute({ a : at_time(a,time_next, True) for a in self.V_next} )

        return f2

    def detime(self, formula : BoolFormula,
               time : int) -> BoolFormula :
        """
        Given a formula containing the variables V "at time i" returns a formula
        containing variables at "no time"

        Example
        Assume V = {a,b,c}, calling detime((a_at_2 & b_at_2 & c_at_2), 2) would
        return "(a & b & c)".
        """

        subs = {self.retime(v,time) : v for v in self.V }
        return formula.substitute(subs)


    def add_invar(self, formula : BoolFormula):
        self.init = And(self.init, formula)
        self.trans = And(self.trans, formula)

        rename = { v : get_next(v) for v in self.V}
        formula_next = formula.substitute(rename)
        self.trans = And(self.trans, formula_next)


def from_vmt(instream, env = None, rename_from_nuxmv=True) -> Tuple[SymbolicTS,list[BoolFormula],list[Any]] :
    """
    Parse a transition system expressed as VMT from a stream.

    Here, we'll only care for Boolean types.
    """

    if env is None:
        env = get_env()

    parser = SmtLibParser(env)
    script = parser.get_script(instream)

    # read all state variables
    state_vars = []
    state_vars = script.annotations.all_annotated_formulae("next")

    rename_map = {}

    for s in state_vars:
        next_vars_str_list = script.annotations.annotations(s)["next"]
        assert((not next_vars_str_list is None) and len(next_vars_str_list) == 1)
        next_var_str = next(iter(next_vars_str_list))
        next_var_vmt = Symbol(next_var_str, s.symbol_type())

        if rename_from_nuxmv and s.symbol_name().endswith("__AT0"):
            new_name = s.symbol_name()[:-len("__AT0")]
            s_new = Symbol(new_name, s.symbol_type())

            assert s_new in get_env().formula_manager
            assert s in get_env().formula_manager
            rename_map[s] = s_new

            s = s_new

        next_var = get_next(s)

        rename_map[next_var_vmt] = next_var


    # TODO: fix input variables (that may not have a next in the vmt format?)
    def get_formula(script, label):
        formula = script.annotations.all_annotated_formulae(label)
        assert((not formula is None) and len(formula) == 1)
        return next(iter(formula))

    init_f = get_formula(script, "init").substitute(rename_map)
    trans_f = get_formula(script, "trans").substitute(rename_map)

    # Read the invariants
    invar_props = []
    for inv_vmt in script.annotations.all_annotated_formulae("invar-property"):
        invar_props.append(inv_vmt.substitute(rename_map))

    state_vars = [v.substitute(rename_map) for v in state_vars]
    return (SymbolicTS(state_vars, init_f, trans_f),
            invar_props, [])

