"""
Wrap the bdd functions from PySMT.

"""


from pysmt.shortcuts import *
from pysmt.logics import BOOL

try:
    import repycudd
except ImportError:
    print("Missing BDD package")

import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    TypeAlias = Any  # Fallback for older versions

BDD : TypeAlias = repycudd.DdNode

class BDDWrapper:
    def __init__(self):
        self.bdd_solver = Solver(logic=BOOL, name='bdd')
        self.bdd_converter = self.bdd_solver.converter
        self.dd = self.bdd_converter.ddmanager

    def get_bdd(self, formula):
        """
        Returns the BDD of a formula
        """
        return self.bdd_converter.convert(formula)

    def bdd_true(self):
        return self.bdd_converter.convert(TRUE())

    def bdd_false(self):
        return self.bdd_converter.convert(FALSE())

    def get_formula(self, bdd):
        return self.bdd_converter.back(bdd)

    def And(self, bdd1, bdd2):
        """
        Returns the And of two bdds
        """
        return self.dd.And(bdd1, bdd2)

    def Or(self, bdd1, bdd2):
        """
        Returns the Or of two bdds
        """
        return self.dd.Or(bdd1, bdd2)

    def Not(self, bdd1):
        """
        Returns the Not of two bdds
        """
        return self.dd.Not(bdd1)

    def Rename(self, bdd, maps):
        """
        Returns the bdd where the variables in maps have been renamed.

        Note: maps still contains PySMT formulae and *not BDDs*.
        """

        permut_len = self.dd.ReadSize()
        permut = repycudd.IntArray(permut_len)
        for i in range(permut_len):
            permut[i] = i;

        for key,value in maps.items():
            # make sure the vars are in dd
            key_bdd = self.bdd_converter.convert(key)
            key_index = key_bdd.NodeReadIndex()

            value_bdd =  self.bdd_converter.convert(value)
            value_index = value_bdd.NodeReadIndex()

            permut[key_index] = value_index

        renamed_bdd = self.dd.Permute(bdd, permut)

        return renamed_bdd

    def Elim(self, vars_list, bdd):
        """
        Returns the bdd computing the existential elimination
        of the variables from  vars_list

        Note: vars_list still contains PySMT formulae, and not Bdds.
        """
        indices = repycudd.IntArray(len(vars_list))
        i = -1
        for v in vars_list:
            i += 1
            indices[i] = self.bdd_converter.convert(v).NodeReadIndex()
        cube = self.dd.IndicesToCube(indices, len(vars_list))

        res = self.dd.ExistAbstract(bdd, cube)
        return res

