# Computation Tree Temporal Logic
#
# Code adapted from the PySMT distribution.
#

from typing import Any, Set, Collection, List

import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    TypeAlias = Any  # Fallback for older versions


from functools import partial

from pysmt.shortcuts import *
from pysmt.fnode import FNode
from pysmt.operators import new_node_type
import pysmt.environment
import pysmt.formula
from pysmt.type_checker import SimpleTypeChecker
from pysmt.oracles import FreeVarsOracle

import pysmt.printers
from pysmt.walkers.generic import handles
from pysmt.walkers import IdentityDagWalker
from pysmt.environment import Environment, pop_env, get_env
from pysmt.environment import push_env as pysmt_push_env
import pysmt.operators as op
from pysmt.walkers import DagWalker
from pysmt.shortcuts import Symbol, get_env
from pysmt.simplifier import Simplifier
from pysmt.shortcuts import Symbol, Not, get_env



# Ctl type (warning must check...)
# 
CtlFormula : TypeAlias = FNode


# New operators for CTL
CTL_EG = new_node_type(node_str="CTL_EG")
CTL_EX = new_node_type(node_str="CTL_EX")
CTL_EF = new_node_type(node_str="CTL_EF")
CTL_EU = new_node_type(node_str="CTL_EU")
CTL_AG = new_node_type(node_str="CTL_AG")
CTL_AX = new_node_type(node_str="CTL_AX")
CTL_AF = new_node_type(node_str="CTL_AF")
CTL_AU = new_node_type(node_str="CTL_AU")

ALL_CTL = (CTL_EG, CTL_EX, CTL_EF, CTL_EU, CTL_AG, CTL_AX, CTL_AF, CTL_AU)


# Extend the FormulaManager
class FormulaManager(pysmt.formula.FormulaManager):
    def EG(self, formula):
        return self.create_node(node_type=CTL_EG, args=(formula,))

    def EX(self, formula):
        return self.create_node(node_type=CTL_EX, args=(formula,))

    def EF(self, formula):
        return self.create_node(node_type=CTL_EF, args=(formula,))

    def EU(self, left, right):
        return self.create_node(node_type=CTL_EU, args=(left, right))

    def AG(self, formula):
        return self.create_node(node_type=CTL_AG, args=(formula,))

    def AX(self, formula):
        return self.create_node(node_type=CTL_AX, args=(formula,))

    def AF(self, formula):
        return self.create_node(node_type=CTL_AF, args=(formula,))

    def AU(self, left, right):
        return self.create_node(node_type=CTL_AU, args=(left, right))


def CTLEG(left):
    return get_env().formula_manager.EG(left)

def CTLEX(left):
    return get_env().formula_manager.EX(left)

def CTLEF(left):
    return get_env().formula_manager.EF(left)

def CTLEU(left, right):
    return get_env().formula_manager.EU(left,right)

def CTLAG(left):
    return get_env().formula_manager.AG(left)

def CTLAX(left):
    return get_env().formula_manager.AX(left)

def CTLAF(left):
    return get_env().formula_manager.AF(left)

def CTLAU(left, right):
    return get_env().formula_manager.AU(left,right)

def is_ctl(self):
    return self.node_type() in ALL_CTL

def is_ctl_EG(self):
    return self.node_type() == CTL_EG
def is_ctl_EX(self):
    return self.node_type() == CTL_EX
def is_ctl_EF(self):
    return self.node_type() == CTL_EF
def is_ctl_EU(self):
    return self.node_type() == CTL_EU
def is_ctl_AG(self):
    return self.node_type() == CTL_AG
def is_ctl_AX(self):
    return self.node_type() == CTL_AX
def is_ctl_AF(self):
    return self.node_type() == CTL_AF
def is_ctl_AU(self):
    return self.node_type() == CTL_AU

FNode.is_ctl_EG = is_ctl_EG
FNode.is_ctl_EX = is_ctl_EX
FNode.is_ctl_EF = is_ctl_EF
FNode.is_ctl_EU = is_ctl_EU
FNode.is_ctl_AG = is_ctl_AG
FNode.is_ctl_AX = is_ctl_AX
FNode.is_ctl_AF = is_ctl_AF
FNode.is_ctl_AU = is_ctl_AU



# Extend the type checker
SimpleTypeChecker.set_handler(SimpleTypeChecker.walk_bool_to_bool, *ALL_CTL)

# Extend the free variable oracle
FreeVarsOracle.set_handler(FreeVarsOracle.walk_simple_args, *ALL_CTL)

def simplify_ctl_eg(self, formula, args, **kwargs): return formula
def simplify_ctl_ex(self, formula, args, **kwargs): return formula
def simplify_ctl_ef(self, formula, args, **kwargs): return formula
def simplify_ctl_eu(self, formula, args, **kwargs): return formula
def simplify_ctl_ag(self, formula, args, **kwargs): return formula
def simplify_ctl_ax(self, formula, args, **kwargs): return formula
def simplify_ctl_af(self, formula, args, **kwargs): return formula
def simplify_ctl_au(self, formula, args, **kwargs): return formula

Simplifier.set_handler(simplify_ctl_eg, CTL_EG)
Simplifier.set_handler(simplify_ctl_ex, CTL_EX)
Simplifier.set_handler(simplify_ctl_ef, CTL_EF)
Simplifier.set_handler(simplify_ctl_eu, CTL_EU)
Simplifier.set_handler(simplify_ctl_ag, CTL_AG)
Simplifier.set_handler(simplify_ctl_ax, CTL_AX)
Simplifier.set_handler(simplify_ctl_af, CTL_AF)
Simplifier.set_handler(simplify_ctl_au, CTL_AU)

# Extend the printer
class HRPrinter(pysmt.printers.HRPrinter):
    CTL_TYPE_TO_STR = {
        CTL_EG: "EG",
        CTL_EX: "EX",
        CTL_EF: "EF",
        CTL_AG: "AG",
        CTL_AX: "AX",
        CTL_AF: "AF" }

    @handles(CTL_EU, CTL_AU)
    def walk_ctl_u(self, formula):
        node_type = formula.node_type()
        path_quantifier = "A" if node_type == CTL_AU else "E"

        self.stream.write("(")
        yield formula.arg(0)
        self.stream.write(" %sU " % path_quantifier)
        yield formula.arg(1)
        self.stream.write(")")

        return self.walk_nary(formula, " U ")

    @handles(CTL_EG, CTL_EX, CTL_EF, CTL_AG, CTL_AX, CTL_AF)
    def walk_ctl(self, formula):
        node_type = formula.node_type()
        op_symbol = self.CTL_TYPE_TO_STR[node_type]
        self.stream.write("(%s " % op_symbol)
        yield formula.arg(0)
        self.stream.write(")")
# EOC HRPrinter

# Extend the serializer
class HRSerializer(pysmt.printers.HRSerializer):
    PrinterClass = HRPrinter
# EOC HRSerialize




# Extend DAG Walker
def walk_ctl_eg(self, formula, args, **kwargs): return self.mgr.EG(args[0])
def walk_ctl_ex(self, formula, args, **kwargs): return self.mgr.EX(args[0])
def walk_ctl_ef(self, formula, args, **kwargs): return self.mgr.EF(args[0])
def walk_ctl_eu(self, formula, args, **kwargs): return self.mgr.EU(args[0], args[1])
def walk_ctl_ag(self, formula, args, **kwargs): return self.mgr.AG(args[0])
def walk_ctl_ax(self, formula, args, **kwargs): return self.mgr.AX(args[0])
def walk_ctl_af(self, formula, args, **kwargs): return self.mgr.AF(args[0])
def walk_ctl_au(self, formula, args, **kwargs): return self.mgr.AU(args[0], args[1])

IdentityDagWalker.set_handler(walk_ctl_eg, CTL_EG)
IdentityDagWalker.set_handler(walk_ctl_ex, CTL_EX)
IdentityDagWalker.set_handler(walk_ctl_ef, CTL_EF)
IdentityDagWalker.set_handler(walk_ctl_eu, CTL_EU)
IdentityDagWalker.set_handler(walk_ctl_ag, CTL_AG)
IdentityDagWalker.set_handler(walk_ctl_ax, CTL_AX)
IdentityDagWalker.set_handler(walk_ctl_af, CTL_AF)
IdentityDagWalker.set_handler(walk_ctl_au, CTL_AU)
# EOC IdentityDagWalker


# Extend the environment (should be probably used when initializing the package to use...)
class Environment(Environment):
    """Extension of pySMT environment."""
    # Only specify new classes. Classes that have been extended
    # directly do not need to be redefined here (e.g., TypeChecker)
    FormulaManagerClass = FormulaManager
    HRSerializerClass = HRSerializer
# EOC Environment


def push_env(env=None):
    """Overload push_env to default to the new Environment class."""
    if env is None:
        env = Environment()
    return pysmt_push_env(env=env)

def reset_env():
    """Overload reset_env to use the new push_env()."""
    pop_env()
    push_env()
    return get_env()

# Create the default environment
reset_env()

def init_ctl():
    return reset_env()


class SubFormulas(IdentityDagWalker):
    """
    Visits the formula and extract the set of subformulas.

    How to use:
    subformulas = SubFormulas.get_subs(formula)
    """

    def __init__(self, env=None):
        DagWalker.__init__(self, env=env)
        self.mgr = self.env.formula_manager
        self._subformulas = set()

    def walk_bool_constant(self, formula, args, **kwargs):
        self._subformulas.add(formula)
        return True

    def walk_symbol(self, formula, args, **kwargs):
        self._subformulas.add(formula)
        return True

    @handles(op.AND, op.OR, op.IFF, op.NOT, op.IMPLIES,
             CTL_EG, CTL_EX, CTL_EF, CTL_EU, CTL_AG, CTL_AX, CTL_AF, CTL_AU)
    def walk_op(self, formula, args, **kwargs):
        self._subformulas.add(formula)
        return True


def get_sub(formula : CtlFormula, env = None) -> CtlFormula:
    """ Returns the set of all the subformuals of formula
    """

    if env is None:
        env = get_env()

    sf = SubFormulas(env)
    sf._subformulas.add(formula)
    sf.walk(formula)
    return sf._subformulas


