# Linear Temporal Logic
#
# Code adapted from the PySMT distribution.
#

from typing import Any, Set, Collection, List

import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    TypeAlias = Any  # Fallback for older versions


from functools import partial

from pysmt.shortcuts import *
from pysmt.fnode import FNode
from pysmt.operators import new_node_type
import pysmt.environment
import pysmt.formula
from pysmt.type_checker import SimpleTypeChecker
from pysmt.oracles import FreeVarsOracle

import pysmt.printers
from pysmt.walkers.generic import handles
from pysmt.walkers import IdentityDagWalker
from pysmt.environment import Environment, pop_env, get_env
from pysmt.environment import push_env as pysmt_push_env
import pysmt.operators as op
from pysmt.walkers import DagWalker
from pysmt.shortcuts import Symbol, get_env
from pysmt.simplifier import Simplifier
from pysmt.shortcuts import Symbol, Not, get_env



# Ltl type (warning must check...)
# 
LtlFormula : TypeAlias = FNode


# New operators for LTL
LTL_X = new_node_type(node_str="LTL_X")
LTL_F = new_node_type(node_str="LTL_F")
LTL_G = new_node_type(node_str="LTL_G")
LTL_U = new_node_type(node_str="LTL_U")
LTL_R = new_node_type(node_str="LTL_R")
ALL_LTL = (LTL_X, LTL_F, LTL_G, LTL_U, LTL_R)


# Extend the FormulaManager
class FormulaManager(pysmt.formula.FormulaManager):
    """Extension of FormulaManager to handle LTL Operators."""
    def X(self, formula):
        return self.create_node(node_type=LTL_X, args=(formula,))

    def F(self, formula):
        return self.create_node(node_type=LTL_F, args=(formula,))

    def G(self, formula):
        return self.create_node(node_type=LTL_G, args=(formula,))

    def U(self, left, right):
        return self.create_node(node_type=LTL_U, args=(left, right))

    def R(self, left, right):
        return self.create_node(node_type=LTL_R, args=(left, right))


def LTLX(left):
    return get_env().formula_manager.X(left)

def LTLF(left):
    return get_env().formula_manager.F(left)

def LTLG(left):
    return get_env().formula_manager.G(left)

def LTLU(left, right):
    return get_env().formula_manager.U(left, right)

def LTLR(left, right):
    return get_env().formula_manager.R(left, right)


def is_ltl_U(self):
    """ """
    return self.node_type() == LTL_U

def is_ltl_R(self):
    """ """
    return self.node_type() == LTL_R

def is_ltl_X(self):
    """ """
    return self.node_type() == LTL_X

def is_ltl_F(self):
    """ """
    return self.node_type() == LTL_F

def is_ltl_G(self):
    """ """
    return self.node_type() == LTL_G


FNode.is_ltl_U = is_ltl_U
FNode.is_ltl_R = is_ltl_R
FNode.is_ltl_X = is_ltl_X
FNode.is_ltl_F = is_ltl_F
FNode.is_ltl_G = is_ltl_G



# Extend the type checker
SimpleTypeChecker.set_handler(SimpleTypeChecker.walk_bool_to_bool, *ALL_LTL)

# Extend the free variable oracle
FreeVarsOracle.set_handler(FreeVarsOracle.walk_simple_args, *ALL_LTL)

# Extend simplifier
def simplify_ltl_x(self, formula, args, **kwargs): return formula
def simplify_ltl_g(self, formula, args, **kwargs): return formula
def simplify_ltl_f(self, formula, args, **kwargs): return formula
def simplify_ltl_u(self, formula, args, **kwargs): return formula
def simplify_ltl_r(self, formula, args, **kwargs): return formula

Simplifier.set_handler(simplify_ltl_x, LTL_X)
Simplifier.set_handler(simplify_ltl_g, LTL_G)
Simplifier.set_handler(simplify_ltl_f, LTL_F)
Simplifier.set_handler(simplify_ltl_u, LTL_U)
Simplifier.set_handler(simplify_ltl_r, LTL_R)

# Extend the printer
class HRPrinter(pysmt.printers.HRPrinter):
    def walk_ltl_u(self, formula):
        return self.walk_nary(formula, " U ")

    def walk_ltl_r(self, formula):
        return self.walk_nary(formula, " R ")

    @handles(LTL_X, LTL_F, LTL_G)
    def walk_ltl(self, formula):
        node_type = formula.node_type()
        LTL_TYPE_TO_STR = { LTL_X: "X", LTL_F: "F", LTL_G: "G"}
        op_symbol = LTL_TYPE_TO_STR[node_type]
        self.stream.write("(%s " % op_symbol)
        yield formula.arg(0)
        self.stream.write(")")
# EOC HRPrinter

# Extend the serializer
class HRSerializer(pysmt.printers.HRSerializer):
    PrinterClass = HRPrinter
# EOC HRSerialize




# Extend DAG Walker
def walk_ltl_x(self, formula, args, **kwargs): return self.mgr.X(args[0])
def walk_ltl_u(self, formula, args, **kwargs): return self.mgr.U(args[0], args[1])
def walk_ltl_r(self, formula, args, **kwargs): return self.mgr.R(args[0], args[1])
def walk_ltl_f(self, formula, args, **kwargs): return self.mgr.F(args[0])
def walk_ltl_g(self, formula, args, **kwargs): return self.mgr.G(args[0])
IdentityDagWalker.set_handler(walk_ltl_x, LTL_X)
IdentityDagWalker.set_handler(walk_ltl_u, LTL_U)
IdentityDagWalker.set_handler(walk_ltl_r, LTL_R)
IdentityDagWalker.set_handler(walk_ltl_f, LTL_F)
IdentityDagWalker.set_handler(walk_ltl_g, LTL_G)
# EOC IdentityDagWalker


# Extend the environment (should be probably used when initializing the package to use...)
class Environment(Environment):
    """Extension of pySMT environment."""
    # Only specify new classes. Classes that have been extended
    # directly do not need to be redefined here (e.g., TypeChecker)
    FormulaManagerClass = FormulaManager
    HRSerializerClass = HRSerializer
# EOC Environment


def push_env(env=None):
    """Overload push_env to default to the new Environment class."""
    if env is None:
        env = Environment()
    return pysmt_push_env(env=env)

def reset_env():
    """Overload reset_env to use the new push_env()."""
    pop_env()
    push_env()
    return get_env()

# Create the default environment
reset_env()

def init_ltl():
    return reset_env()


class LtlNormalizer(IdentityDagWalker):
    """
    Rewrites the LTL formula in an equivalent formula without |, G, and F operator.
    """

    def __init__(self, env=None, invalidate_memoization=None):
        IdentityDagWalker.__init__(self,
                                   env=env,
                                   invalidate_memoization=invalidate_memoization)
        self.mgr = self.env.formula_manager

    def walk_not(self, formula, args, **kwargs):
        """ Remove double negation """
        if args[0].is_not():
            args[0].args()[0]
        else:
            return self.mgr.Not(args[0])

    def walk_or(self, formula, args, **kwargs):
        # unflatten the tree
        and_sequence = None
        for f in args:
            neg_f = self.mgr.Not(f)
            if and_sequence is None:
                and_sequence = neg_f
            else:
                and_sequence = self.mgr.And(and_sequence, neg_f)

        return self.mgr.Not(and_sequence)

    def walk_and(self, formula, args, **kwargs):
        # unflatten the tree
        and_sequence = None
        for f in args:
            if and_sequence is None:
                and_sequence = f
            else:
                and_sequence = self.mgr.And(and_sequence, f)

        return and_sequence


    def walk_implies(self, formula, args, **kwargs):
        # ! (a & !b)
        return self.mgr.Not(
            self.mgr.And(args[0],
                         self.mgr.Not(args[1]))
        )

    def walk_ltl_x(self, formula, args, **kwargs):
        return self.mgr.X(args[0])

    def walk_iff(self, formula, args, **kwargs):
        # a <-> b
        # a -> b & b -> a
        # (!a | b) & (!b | a)
        # ( (! (a & !b) & !(b & !a))
        p1 = self.mgr.Not(self.mgr.And(args[0], self.mgr.Not(args[1])))
        p2 = self.mgr.Not(self.mgr.And(self.mgr.Not(args[0]), args[1]))
        iff_res = self.mgr.And(p1, p2)

        return iff_res

    def walk_ltl_f(self, formula, args, **kwargs):
        return self.mgr.U(self.mgr.TRUE(), args[0])

    def walk_ltl_g(self, formula, args, **kwargs):
        return self.mgr.Not(self.mgr.U(self.mgr.TRUE(), self.mgr.Not(args[0])))


class SubFormulas(IdentityDagWalker):
    """
    Visits the formula and extract the set of subformulas.

    How to use:
    subformulas = SubFormulas.get_subs(formula)
    """

    def __init__(self, env=None):
        DagWalker.__init__(self, env=env)
        self.mgr = self.env.formula_manager
        self._subformulas = set()

    def walk_bool_constant(self, formula, args, **kwargs):
        self._subformulas.add(formula)
        return True

    def walk_symbol(self, formula, args, **kwargs):
        self._subformulas.add(formula)
        return True

    @handles(op.AND, op.OR, op.IFF, op.NOT, op.IMPLIES, LTL_X, LTL_F, LTL_G, LTL_U, LTL_R)
    def walk_op(self, formula, args, **kwargs):
        self._subformulas.add(formula)
        return True


def normalize_ltl(formula : LtlFormula, env = None) -> LtlFormula:
    """ Normalize a LTL formula using only !, &, and U
    """
    if env is None:
        env = get_env()

    n = LtlNormalizer(env)
    return n.walk(formula)


def get_sub(formula : LtlFormula, env = None) -> LtlFormula:
    """ Returns the set of all the subformuals of formula
    """

    if env is None:
        env = get_env()

    sf = SubFormulas(env)
    sf._subformulas.add(formula)
    sf.walk(formula)
    return sf._subformulas


