import ply.lex as lex
import ply.yacc as yacc

from mc308.logics.ctl import reset_env # use CTL

from pysmt.shortcuts import get_env


# Lexer


reserved = {
    'True' : 'TRUE',
    'False' : 'FALSE',
    'EG' : 'EG',
    'EX' : 'EX',
    'EF' : 'EF',
    'EU' : 'EU',
    'AG' : 'AG',
    'AX' : 'AX',
    'AF' : 'AF',
    'AU' : 'AU',
}

# List of token names
tokens = (
    'IDENTIFIER',
    'AND', 'OR', 'NOT', 'IMPLIES', 'IFF',
    'LPAREN', 'RPAREN'
) + tuple(reserved.values())


#t_TRUE = r'True'
#t_FALSE = r'False'
t_AND = r'\&'
t_OR = r'\|'
t_NOT = r'\!'
t_IMPLIES = r'->'
t_IFF = r'<->'
t_LPAREN = r'\('
t_RPAREN = r'\)'

# Regular expressions for the tokens

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'IDENTIFIER')    # Check for reserved words
    return t

# Ignore whitespace
t_ignore = ' \t'

# from lower to higher
precedence = (
    ('right', 'IMPLIES', 'IFF'),
    ('right', 'AND', 'OR'),
    ('right','EU','AU'),
    ('right', 'EF','AF','EG','AG'),
    ('right','EX','AX'),
    ('right', 'NOT'),
)

# Error handling rule
def t_error(t):
    raise Exception(f"Lexer error - Illegal character '{t.value[0]}'")
    t.lexer.skip(1)

# Build the lexer
ctl_lexer = lex.lex()

# Parser

# Parsing rules
def p_expression_true(p):
    'expression : TRUE'
    p[0] = p.parser.context.mgr.TRUE()

def p_expression_false(p):
    'expression : FALSE'
    p[0] = p.parser.context.mgr.FALSE()

def p_expression_identifier(p):
    'expression : IDENTIFIER'
    p[0] = p.parser.context.mgr.Symbol(p[1])

def p_expression_not(p):
    'expression : NOT expression'
    p[0] = p.parser.context.mgr.Not(p[2])

def p_expression_and(p):
    'expression : expression AND expression'
    p[0] = p.parser.context.mgr.And(p[1],p[3])

def p_expression_or(p):
    'expression : expression OR expression'
    p[0] = p.parser.context.mgr.Or(p[1],p[3])

def p_expression_implies(p):
    'expression : expression IMPLIES expression'
    p[0] = p.parser.context.mgr.Implies(p[1],p[3])

def p_expression_iff(p):
    'expression : expression IFF expression'
    p[0] = p.parser.context.mgr.Iff(p[1],p[3])

def p_expression_eu(p):
    'expression : expression EU expression'
    p[0] = p.parser.context.mgr.EU(p[1],p[3])

def p_expression_au(p):
    'expression : expression AU expression'
    p[0] = p.parser.context.mgr.AU(p[1],p[3])

def p_expression_ef(p):
    'expression : EF expression'
    p[0] = p.parser.context.mgr.EF(p[2])

def p_expression_af(p):
    'expression : AF expression'
    p[0] = p.parser.context.mgr.AF(p[2])

def p_expression_eg(p):
    'expression : EG expression'
    p[0] = p.parser.context.mgr.EG(p[2])

def p_expression_ag(p):
    'expression : AG expression'
    p[0] = p.parser.context.mgr.AG(p[2])

def p_expression_ex(p):
    'expression : EX expression'
    p[0] = p.parser.context.mgr.EX(p[2])

def p_expression_ax(p):
    'expression : AX expression'
    p[0] = p.parser.context.mgr.AX(p[2])

def p_expression_group(p):
    'expression : LPAREN expression RPAREN'
    p[0] = p[2]

# Error rule for syntax errors
def p_error(p):
    raise Exception("Syntax error in input!")

# Build the parser
ctl_parser = yacc.yacc()

class CTLContext:
    def __init__(self, env):
        self.env = env
        self.mgr = env.formula_manager

class CTLParser:
    def __init__(self, env):
        self.lexer = ctl_lexer.clone()
        self.parser = ctl_parser
        self.parser.context = CTLContext(env)


    def parse(self, text):
        return self.parser.parse(text, lexer=self.lexer)


