import ply.lex as lex
import ply.yacc as yacc

from mc308.logics.ltl import reset_env # use LTL

from pysmt.shortcuts import get_env


# Lexer

reserved = {
   'True' : 'TRUE',
   'False' : 'FALSE',
   'U' : 'UNTIL',
   'R' : 'RELEASE',
   'F' : 'FINALLY',
   'G' : 'GLOBALLY',
   'X' : 'NEXT',
}

# List of token names
tokens = (
    'IDENTIFIER',
    'AND', 'OR', 'NOT', 'IMPLIES', 'IFF',
    'LPAREN', 'RPAREN'
) + tuple(reserved.values())


#t_TRUE = r'True'
#t_FALSE = r'False'
t_AND = r'\&'
t_OR = r'\|'
t_NOT = r'\!'
t_IMPLIES = r'->'
t_IFF = r'<->'
t_LPAREN = r'\('
t_RPAREN = r'\)'

# Regular expressions for the tokens

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'IDENTIFIER')    # Check for reserved words
    return t

# Ignore whitespace
t_ignore = ' \t'

# from lower to higher
precedence = (
    ('right', 'IMPLIES', 'IFF'),
    ('right', 'AND', 'OR'),
    ('right', 'UNTIL', 'RELEASE'),
    ('right', 'FINALLY', "GLOBALLY"),
    ('right', 'NOT'),
)

# Error handling rule
def t_error(t):
    raise Exception(f"Lexer error - Illegal character '{t.value[0]}'")
    t.lexer.skip(1)

# Build the lexer
ltl_lexer = lex.lex()

# Parser

# Parsing rules
def p_expression_true(p):
    'expression : TRUE'
    p[0] = p.parser.context.mgr.TRUE()

def p_expression_false(p):
    'expression : FALSE'
    p[0] = p.parser.context.mgr.FALSE()

def p_expression_identifier(p):
    'expression : IDENTIFIER'
    p[0] = p.parser.context.mgr.Symbol(p[1])

def p_expression_not(p):
    'expression : NOT expression'
    p[0] = p.parser.context.mgr.Not(p[2])

def p_expression_and(p):
    'expression : expression AND expression'
    p[0] = p.parser.context.mgr.And(p[1],p[3])

def p_expression_or(p):
    'expression : expression OR expression'
    p[0] = p.parser.context.mgr.Or(p[1],p[3])

def p_expression_implies(p):
    'expression : expression IMPLIES expression'
    p[0] = p.parser.context.mgr.Implies(p[1],p[3])

def p_expression_iff(p):
    'expression : expression IFF expression'
    p[0] = p.parser.context.mgr.Iff(p[1],p[3])

def p_expression_until(p):
    'expression : expression UNTIL expression'
    p[0] = p.parser.context.mgr.U(p[1],p[3])

def p_expression_release(p):
    'expression : expression RELEASE expression'
    p[0] = p.parser.context.mgr.R(p[1],p[3])

def p_expression_finally(p):
    'expression : FINALLY expression'
    p[0] = p.parser.context.mgr.F(p[2])

def p_expression_globally(p):
    'expression : GLOBALLY expression'
    p[0] = p.parser.context.mgr.G(p[2])

def p_expression_next(p):
    'expression : NEXT expression'
    p[0] = p.parser.context.mgr.X(p[2])

def p_expression_group(p):
    'expression : LPAREN expression RPAREN'
    p[0] = p[2]

# Error rule for syntax errors
def p_error(p):
    raise Exception("Syntax error in input!")

# Build the parser
ltl_parser = yacc.yacc()

class LTLContext:
    def __init__(self, env):
        self.env = env
        self.mgr = env.formula_manager

class LTLParser:
    def __init__(self, env):
        self.lexer = ltl_lexer.clone()
        self.parser = ltl_parser
        self.parser.context = LTLContext(env)


    def parse(self, text):
        return self.parser.parse(text, lexer=self.lexer)


